import re
from collections import Counter

import aiohttp

from ._parsers import (
    parse_comments,
    parse_subreddits,
    parse_subreddit_wiki_page,
    parse_posts,
    parse_users,
)
from .api import Api, TIMEFRAME, SORT_CRITERION, LISTING, TIME_FORMAT

api = Api()


class User:
    """Represents a Reddit user and provides methods for getting data from the specified user."""

    def __init__(self, username: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a User instance for getting profile, posts and comments data from the specified user.

        :param username: Username of the user to get data from.
        :type username: str
        :param time_format: Determines the format of the output. Use "concise" for a human-readable
                        time difference, or "locale" for a localized datetime string. Defaults to "locale".
        :type time_format: Literal["concise", "locale"]
        """
        self._username = username
        self._time_format = time_format

    async def profile(self, session: aiohttp.ClientSession) -> dict:
        """
        Returns a user's profile data.

        :param session: aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A dictionary containing user profile data.
        :rtype: dict
        """
        raw_user: dict = await api.get_profile(
            profile_source=self._username, profile_type="user", session=session
        )
        user_profile: dict = parse_users(raw_user, time_format=self._time_format)

        return user_profile

    async def posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns a user's posts.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        raw_posts: list = await api.get_posts(
            posts_source=self._username,
            posts_type="user_posts",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        user_posts: list[dict] = parse_posts(raw_posts, time_format=self._time_format)

        return user_posts

    async def comments(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns a user's comments.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of comments to return.
        :type limit: int
        :param sort: Sort criterion for the comments.
        :type sort: str
        :param timeframe: Timeframe from which tyo get comments.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a comment.
        :rtype: list[dict]
        """
        raw_comments: list = await api.get_posts(
            posts_source=self._username,
            posts_type="user_comments",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        user_comments: list[dict] = parse_comments(
            raw_comments, time_format=self._time_format
        )

        return user_comments

    async def overview(self, limit: int, session: aiohttp.ClientSession) -> list[dict]:
        """
        Returns a user's most recent comments.

        :param limit: Maximum number of comments to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing data about a recent comment.
        :rtype: list[dict]
        """
        raw_comments: list = await api.get_posts(
            posts_source=self._username,
            posts_type="user_overview",
            limit=limit,
            session=session,
        )
        user_overview: list[dict] = parse_comments(
            raw_comments, time_format=self._time_format
        )

        return user_overview

    async def search_posts(
        self,
        session: aiohttp.ClientSession,
        keyword: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns a user's posts that contain the specified keywords.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param keyword: Keyword to search for in posts.
        :type keyword: str
        :param limit: Maximum number of posts to search from.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        user_posts: list = await api.get_posts(
            posts_type="user_posts",
            posts_source=self._username,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        found_posts: list = []
        pattern = re.compile(re.escape(keyword), re.IGNORECASE)
        for post in user_posts:
            post_data: dict = post.get("data")
            if pattern.search(post_data.get("title")) or pattern.search(
                post_data.get("selftext")
            ):
                found_posts.append(post)

        return parse_posts(found_posts, time_format=self._time_format)

    async def search_comments(
        self,
        session: aiohttp.ClientSession,
        keyword: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns a user's comments that contain the specified keyword.

        :param session: Aiohttp session to use for the request.
        :param keyword: Keyword to search for in comments.
        :type keyword: str
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of comments to search from.
        :type limit: int
        :param sort: Sort criterion for the comments.
        :type sort: str
        :param timeframe: Timeframe from which to get comments.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a comment.
        :rtype: list[dict]
        """
        user_comments: list = await api.get_posts(
            posts_source=self._username,
            posts_type="user_comments",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        found_comments: list = []
        pattern = re.compile(re.escape(keyword), re.IGNORECASE)
        for comment in user_comments:
            if pattern.search(comment.get("data").get("body")):
                found_comments.append(comment)

        return parse_comments(found_comments, time_format=self._time_format)

    async def moderated_subreddits(self, session: aiohttp.ClientSession) -> list[dict]:
        """
        Returns subreddits moderated by the user.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing preview data of a Community.
        :rtype: list[dict]
        """
        raw_preview_subreddits: dict = await api.get_data(
            endpoint=f"{api.base_reddit_endpoint}/user/{self._username}/moderated_subreddits.json",
            session=session,
        )
        preview_subreddits: list[dict] = parse_subreddits(
            raw_preview_subreddits.get("data"),
            is_preview=True,
            time_format=self._time_format,
        )

        return preview_subreddits

    async def top_subreddits(
        self,
        session: aiohttp.ClientSession,
        top_n: int,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[tuple]:
        """
        Returns a user's top n subreddits based on subreddit frequency in n posts.

        :param top_n: Communities arranging number.
        :type top_n: int
        :param limit: Maximum number of posts to scrape.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: Dictionary of top n subreddits and their counts.
        :rtype: dict
        """
        posts = await api.get_posts(
            posts_type="user_posts",
            posts_source=self._username,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )

        if posts:
            # Extract subreddit names
            subreddits = [post.get("data", {}).get("subreddit") for post in posts]

            # Count the occurrences of each subreddit
            subreddit_counts = Counter(subreddits)

            return subreddit_counts.most_common(top_n)

        return []


class Search:
    """Represents Readit search functionality and provides methods for getting search results from different entities"""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        self._time_format = time_format

    async def users(
        self,
        query: str,
        limit: int,
        session: aiohttp.ClientSession,
    ) -> list[dict]:
        """
        Search users.

        :param query: Search query.
        :type query: str
        :param limit: Maximum number of search results to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        """
        search_users: list = await api.get_search_results(
            query=query, search_type="users", limit=limit, session=session
        )
        users_results: list[dict] = parse_users(
            search_users, time_format=self._time_format
        )

        return users_results

    async def subreddits(
        self, query: str, limit: int, session: aiohttp.ClientSession
    ) -> list[dict]:
        """
        Search subreddits.

        :param query: Search query.
        :type query: str
        :param limit: Maximum number of search results to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        """
        search_subreddits: list = await api.get_search_results(
            query=query, search_type="subreddits", limit=limit, session=session
        )
        subreddits_results: list[dict] = parse_subreddits(
            search_subreddits, time_format=self._time_format
        )

        return subreddits_results

    async def posts(
        self,
        query: str,
        limit: int,
        session: aiohttp.ClientSession,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns posts.

        :param query: Search query.
        :type query: str
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        search_posts: list = await api.get_search_results(
            query=query,
            search_type="posts",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        posts_results: list[dict] = parse_posts(
            search_posts, time_format=self._time_format
        )

        return posts_results


class Subreddit:
    """Represents a Reddit Community (Subreddit) and provides methods for getting data from the specified subreddit."""

    def __init__(self, subreddit: str, time_format: TIME_FORMAT = "locale"):
        """
        Initialises a RedditCommunity instance for getting profile and posts from the specified subreddit.

        :param subreddit: Name of the subreddit to get data from.
        :type subreddit: str
        :param time_format: Determines the format of the output. Use "concise" for a human-readable
                        time difference, or "locale" for a localized datetime string. Defaults to "locale".
        :type time_format: Literal["concise", "locale"]
        """
        self._subreddit = subreddit
        self._time_format = time_format

    async def profile(self, session: aiohttp.ClientSession) -> dict:
        """
        Returns a subreddit's profile data.

        :param session: aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A dictionary containing subreddit profile data.
        :rtype: Community
        """
        raw_subreddit: dict = await api.get_profile(
            profile_type="subreddit",
            profile_source=self._subreddit,
            session=session,
        )
        subreddit_profile: dict = parse_subreddits(
            raw_subreddit, time_format=self._time_format
        )

        return subreddit_profile

    async def wiki_pages(self, session: aiohttp.ClientSession) -> list[str]:
        """
        Return a subreddit's wiki pages.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of strings, each representing a wiki page.
        :rtype: list[str]
        """
        pages: dict = await api.get_data(
            endpoint=f"{api.subreddit_data_endpoint}/{self._subreddit}/wiki/pages.json",
            session=session,
        )

        return pages.get("data")

    async def wiki_page(self, page: str, session: aiohttp.ClientSession) -> dict:
        """
        Return a subreddit's specified wiki page data.

        :param page: Wiki page to get data from.
        :type page: str
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of strings, each representing a wiki page.
        :rtype: list[str]
        """
        raw_page: dict = await api.get_data(
            endpoint=f"{api.subreddit_data_endpoint}/{self._subreddit}/wiki/{page}.json",
            session=session,
        )
        wiki_page: dict = parse_subreddit_wiki_page(
            wiki_page=raw_page, time_format=self._time_format
        )

        return wiki_page

    async def posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns a subreddit's posts.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        raw_posts: list = await api.get_posts(
            posts_type="subreddit_posts",
            posts_source=self._subreddit,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        subreddit_posts: list[dict] = parse_posts(
            raw_posts, time_format=self._time_format
        )

        return subreddit_posts

    async def search(
        self,
        session: aiohttp.ClientSession,
        keyword: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns posts that contain a specified keyword from a subreddit.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param keyword: Keyword to search for in posts.
        :type keyword: str
        :param limit: Maximum number of posts to search from.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        all_posts: list = await api.get_posts(
            posts_type="subreddit_posts",
            posts_source=self._subreddit,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        found_posts: list = []
        pattern = re.compile(re.escape(keyword), re.IGNORECASE)
        for post in all_posts:
            post_data: dict = post.get("data")
            if pattern.search(post_data.get("title")) or pattern.search(
                post_data.get("selftext")
            ):
                found_posts.append(post)

        return parse_posts(found_posts, time_format=self._time_format)


class Subreddits:
    """Represents Reddit subreddits and provides methods for getting related data."""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        self._time_format = time_format

    async def all(self, limit: int, session: aiohttp.ClientSession) -> list[dict]:
        """
        Return all subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing data about a subreddit.
        :rtype: list[dict]

        Note
        ----
            *in Morphius' voice* "the only limitation you have at this point is the matrix's rate limit."
        """
        raw_subreddits: list = await api.get_subreddits(
            subreddits_type="all", limit=limit, session=session
        )
        all_subreddits: list[dict] = parse_subreddits(
            raw_subreddits, is_preview=True, time_format=self._time_format
        )

        return all_subreddits

    async def default(self, limit: int, session: aiohttp.ClientSession) -> list[dict]:
        """
        Return default subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing data about a subreddit.
        :rtype: list[dict]
        """
        raw_subreddits: list = await api.get_subreddits(
            subreddits_type="default", limit=limit, session=session
        )
        default_subreddits: list[dict] = parse_subreddits(
            raw_subreddits, is_preview=True, time_format=self._time_format
        )

        return default_subreddits

    async def new(self, limit: int, session: aiohttp.ClientSession) -> list[dict]:
        """
        Return new subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing data about a subreddit.
        :rtype: list[dict]
        """
        raw_subreddits: list = await api.get_subreddits(
            subreddits_type="new", limit=limit, session=session
        )
        new_subreddits: list[dict] = parse_subreddits(
            raw_subreddits, is_preview=True, time_format=self._time_format
        )

        return new_subreddits

    async def popular(self, limit: int, session: aiohttp.ClientSession) -> list[dict]:
        """
        Return popular subreddits.

        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :return: A list of dictionaries, each containing data about a subreddit.
        :rtype: list[dict]
        """
        raw_subreddits: list = await api.get_subreddits(
            subreddits_type="popular", limit=limit, session=session
        )
        popular_subreddits: list[dict] = parse_subreddits(
            raw_subreddits, is_preview=True, time_format=self._time_format
        )

        return popular_subreddits


class Post:
    """Represents a Reddit post and provides method(s) for getting data from the specified post."""

    def __init__(
        self, post_id: str, subreddit: str, time_format: TIME_FORMAT = "locale"
    ):
        self._post_id = post_id
        self._subreddit = subreddit
        self._time_format = time_format

    async def profile(self, session: aiohttp.ClientSession) -> dict:
        post_data: dict = await api.get_data(
            endpoint=f"{api.base_reddit_endpoint}/r/{self._subreddit}/comments/{self._post_id}.json",
            session=session,
        )

        raw_post: dict = (
            post_data[0].get("data", {}).get("children", [])[0].get("data", {})
        )
        if raw_post:
            return parse_posts(data=raw_post, time_format=self._time_format)

    async def comments(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
    ) -> list[dict]:
        post_data: dict = await api.get_data(
            endpoint=f"{api.base_reddit_endpoint}/r/{self._subreddit}/comments/"
            f"{self._post_id}.json?sort={sort}&limit={limit}",
            session=session,
        )

        raw_comments: list = post_data[1].get("data", {}).get("children", [])

        if raw_comments:
            return parse_comments(comments=raw_comments, time_format=self._time_format)


class Posts:
    """Represents Reddit posts and provides methods for getting posts from various sources."""

    def __init__(self, time_format: TIME_FORMAT = "locale"):
        self._time_format = time_format

    async def listing(
        self,
        session: aiohttp.ClientSession,
        listings_name: LISTING,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns posts from a specified listing.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param listings_name: Listing to get posts from.
        :type listings_name: str
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        listing_posts: list = await api.get_posts(
            posts_type="listing_posts",
            posts_source=listings_name,
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        if listing_posts:
            return parse_posts(listing_posts, time_format=self._time_format)

    async def new(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
    ) -> list[dict]:
        """
        Returns new posts.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        raw_posts: list = await api.get_posts(
            posts_type="new_posts",
            limit=limit,
            sort=sort,
            session=session,
        )
        new_posts: list[dict] = parse_posts(raw_posts, time_format=self._time_format)

        return new_posts

    async def front_page(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Returns posts from the Reddit front-page.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession.
        :param limit: Maximum number of posts to return.
        :type limit: int
        :param sort: Sort criterion for the posts.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        :return: A list of dictionaries, each containing data about a post.
        :rtype: list[dict]
        """
        raw_posts: list = await api.get_posts(
            posts_type="front_page_posts",
            limit=limit,
            sort=sort,
            timeframe=timeframe,
            session=session,
        )
        front_page_posts: list = parse_posts(raw_posts, time_format=self._time_format)

        return front_page_posts



from typing import Union

from ._utilities import timestamp_to_readable
from .api import TIME_FORMAT


def parse_users(
    data: Union[list[dict], dict], time_format: TIME_FORMAT
) -> Union[list[dict], dict]:
    """
    Parses a list/single object of raw user data.

    :param data: A list of dict objects, each containing raw user data.
    :type data: list[dict]
    :param time_format: TIme format for the parsed data.
    :type time_format: Literal[locale, concise]
    :return: A list of parsed dict objects, each containing user data Or a dict object containing user data.
    :rtype: Union[list[dict], dict]
    """

    def build_user(user: dict) -> dict:
        """
        Parses raw user data to get only the needed data.

        :param user: Raw user data.
        :type user: dict
        :return: A dict object containing parsed user data.
        :rtype: dict
        """
        return {
            "name": user.get("name"),
            "id": user.get("id"),
            "avatar_url": user.get("icon_img"),
            "is_verified": user.get("verified"),
            "has_verified_email": user.get("has_verified_email"),
            "is_gold": user.get("is_gold"),
            "is_mod": user.get("is_mod"),
            "is_blocked": user.get("is_blocked"),
            "is_employee": user.get("is_employee"),
            "hidden_from_bots": user.get("hide_from_robots"),
            "accepts_followers": user.get("accept_followers"),
            "comment_karma": user.get("comment_karma"),
            "link_karma": user.get("link_karma"),
            "awardee_karma": user.get("awardee_karma"),
            "total_karma": user.get("total_karma"),
            "subreddit": user.get("subreddit"),
            "created": (
                timestamp_to_readable(
                    timestamp=user.get("created"), time_format=time_format
                )
                if user.get("created")
                else "NaN"
            ),
        }

    converted_data = None
    if isinstance(data, list) and len(data) != 0:
        converted_data = [build_user(user.get("data")) for user in data]

    elif isinstance(data, dict) and "is_employee" in data:
        converted_data = build_user(data)

    return converted_data


def parse_posts(
    data: Union[dict, list], time_format: TIME_FORMAT
) -> Union[list[dict], dict]:
    """
    Parses a list/single object of raw post data.

    :param data: A list of dict objects, each containing raw post data.
    :type data: list[dict]
    :param time_format: TIme format for the parsed data.
    :type time_format: Literal[locale, concise]
    :return: A list of parsed dict objects, each containing post data Or a dict object containing post data.
    :rtype: Union[list[dict], dict]
    """

    def build_post(post: dict) -> dict:
        """
        Parses raw post data to get only the needed data.

        :param post: Raw post data.
        :type post: dict
        :return: A dict object containing parsed post data.
        :rtype: dict
        """
        return {
            "author": post.get("author"),
            "title": post.get("title"),
            "body": post.get("selftext"),
            "id": post.get("id"),
            "subreddit": post.get("subreddit"),
            "subreddit_id": post.get("subreddit_id"),
            "subreddit_type": post.get("subreddit_type"),
            "upvotes": post.get("ups"),
            "upvote_ratio": post.get("upvote_ratio"),
            "downvotes": post.get("downs"),
            "thumbnail": post.get("thumbnail"),
            "gilded": post.get("gilded"),
            "is_nsfw": post.get("over_18"),
            "is_shareable": post.get("is_reddit_media_domain"),
            "is_robot_indexable": post.get("is_robot_indexable"),
            "permalink": post.get("permalink"),
            "is_locked": post.get("locked"),
            "is_archived": post.get("archived"),
            "domain": post.get("domain"),
            "score": post.get("score"),
            "edited": (
                timestamp_to_readable(
                    timestamp=post.get("edited"), time_format=time_format
                )
                if post.get("edited")
                else False
            ),
            "comments": post.get("num_comments"),
            "created": (
                timestamp_to_readable(
                    timestamp=post.get("created"), time_format=time_format
                )
                if post.get("created")
                else "NaN"
            ),
        }

    if isinstance(data, dict):
        return build_post(post=data)

    elif isinstance(data, list):
        return [build_post(post.get("data")) for post in data if post.get("data")]


def parse_comments(comments: list[dict], time_format: TIME_FORMAT) -> list[dict]:
    """
    Parses a list of raw comments data to get only the needed data in each item.

    :param comments: A list of dict objects, each containing raw comment data.
    :type comments: list[dict]
    :param time_format: TIme format for the parsed data.
    :type time_format: Literal[locale, concise]
    :return: A list of parsed dict objects, each containing comments data.
    :rtype: list[dict]
    """
    if len(comments) != 0:
        comments_list: list = []
        for comment in comments:
            comment_data: dict = comment.get("data")
            comments_list.append(
                {
                    "body": comment_data.get("body"),
                    "id": comment_data.get("id"),
                    "author": comment_data.get("author"),
                    "author_is_premium": comment_data.get("author_premium"),
                    "upvotes": comment_data.get("ups"),
                    "downvotes": comment_data.get("downs"),
                    "subreddit": comment_data.get("subreddit_name_prefixed"),
                    "subreddit_type": comment_data.get("subreddit_type"),
                    "post_id": comment_data.get("link_id"),
                    "post_title": comment_data.get("link_title"),
                    "is_nsfw": comment_data.get("over_18"),
                    "is_edited": comment_data.get("edited"),
                    "score": comment_data.get("score"),
                    "hidden_score": comment_data.get("score_hidden"),
                    "gilded": comment_data.get("gilded"),
                    "is_stickied": comment_data.get("stickied"),
                    "is_locked": comment_data.get("locked"),
                    "is_archived": comment_data.get("archived"),
                    "created": (
                        timestamp_to_readable(
                            timestamp=comment.get("created"), time_format=time_format
                        )
                        if comment.get("created")
                        else "NaN"
                    ),
                }
            )

        return comments_list


def parse_subreddits(
    data: Union[list, dict], time_format, is_preview: bool = False
) -> Union[list[dict], dict]:
    """
    Parses a list/single object of raw subreddits data.

    :param data: A list of dict objects, each containing raw subreddit data.
    :type data: list[dict]
    :param time_format: TIme format for the parsed data.
    :type time_format: Literal[locale, concise]
    :param is_preview: A boolean value to determine whether a subreddit is a preview,
        i.e. Its object contains fewer data.
    :return: A list of parsed dict objects, each containing subreddits data Or a dict object containing subreddit data.
    :rtype: Union[list[dict], dict]
    """

    def build_subreddit(
        subreddit: dict,
    ) -> dict:
        """
        Parses raw subreddit data to get only the needed data.

        :param subreddit: Raw subreddit data.
        :type subreddit: dict
        :return: A dict object containing parsed subreddit data.
        :rtype: dict
        """
        if is_preview:
            subreddit_obj = {
                "name": subreddit.get("display_name"),
                "id": subreddit.get("id"),
                "type": subreddit.get("subreddit_type"),
                "icon": subreddit.get("subreddit_icon", "").split("?")[0],
                "subscribers": subreddit.get("subscribers"),
                "whitelist_status": subreddit.get("whitelist_status"),
                "url": subreddit.get("url"),
                "created": timestamp_to_readable(
                    timestamp=subreddit.get("created"), time_format=time_format
                ),
            }
        else:
            subreddit_obj = {
                "name": subreddit.get("display_name"),
                "id": subreddit.get("id"),
                "description": subreddit.get("public_description"),
                "submit_text": subreddit.get("submit_text"),
                "icon": subreddit.get("subreddit_icon", "").split("?")[0],
                "type": subreddit.get("subreddit_type"),
                "subscribers": subreddit.get("subscribers"),
                "current_active_users": subreddit.get("accounts_active"),
                "is_nsfw": subreddit.get("over18"),
                "language": subreddit.get("lang"),
                "whitelist_status": subreddit.get("whitelist_status"),
                "url": subreddit.get("url"),
                "created": (
                    timestamp_to_readable(
                        timestamp=subreddit.get("created"), time_format=time_format
                    )
                    if subreddit.get("created")
                    else "NaN"
                ),
            }

        return subreddit_obj

    subreddit_data = None
    if isinstance(data, list) and len(data) != 0:
        subreddit_data = [
            build_subreddit(subreddit=subreddit.get("data")) for subreddit in data
        ]

    elif isinstance(data, dict) and "subreddit_type" in data:
        subreddit_data = build_subreddit(subreddit=data)

    return subreddit_data


def parse_subreddit_wiki_page(wiki_page: dict, time_format: TIME_FORMAT) -> dict:
    """
    Parses raw subreddit wiki page data to get only the needed data.

    :param wiki_page: Raw wiki page data.
    :type wiki_page: dict
    :param time_format: TIme format for the parsed data.
    :type time_format: Literal[locale, concise]
    :return: A dict object of parsed wiki page data.
    :rtype: dict
    """
    if "revision_id" in wiki_page:
        page_data: dict = wiki_page.get("data")
        user: dict = page_data.get("revision_by").get("data")

        return {
            "revision_id": page_data.get("revision_id"),
            "revision_date": timestamp_to_readable(
                timestamp=page_data.get("revision_date")
            ),
            "content_markdown": page_data.get("content_md"),
            "revised_by": {
                "name": user.get("name"),
                "id": user.get("id"),
                "avatar_url": user.get("icon_img"),
                "is_verified": user.get("verified"),
                "has_verified_email": user.get("has_verified_email"),
                "is_gold": user.get("is_gold"),
                "is_mod": user.get("is_mod"),
                "is_blocked": user.get("is_blocked"),
                "is_employee": user.get("is_employee"),
                "hidden_from_bots": user.get("hide_from_robots"),
                "accepts_followers": user.get("accept_followers"),
                "comment_karma": user.get("comment_karma"),
                "link_karma": user.get("link_karma"),
                "awardee_karma": user.get("awardee_karma"),
                "total_karma": user.get("total_karma"),
                "subreddit": user.get("subreddit"),
                "created": timestamp_to_readable(
                    timestamp=user.get("created"), time_format=time_format
                ),
            },
        }



import argparse
import asyncio
import os
from datetime import datetime
from typing import get_args

import aiohttp
from rich.markdown import Markdown
from rich.tree import Tree
from rich_argparse import RichHelpFormatter

from ._core import Post, Posts, Subreddit, Subreddits, User, Search
from ._utilities import (
    console,
    create_dataframe,
    export_dataframe,
    filename_timestamp,
    pathfinder,
    print_banner,
    show_exported_files,
)
from .api import Api, TIMEFRAME, SORT_CRITERION, LISTING
from .help import Help
from .version import Version


def create_parser() -> argparse.ArgumentParser:
    """
    Creates and configures an argument parser for the command line arguments.

    :return: A configured argparse.ArgumentParser object ready to parse the command line arguments.
    :rtype: argparse.ArgumentParser
    """
    main_parser = argparse.ArgumentParser(
        description=Markdown(Help.summary, style="argparse.text"),
        epilog=Markdown(Help.description),
        formatter_class=RichHelpFormatter,
    )
    subparsers = main_parser.add_subparsers(dest="module", help="module")
    main_parser.add_argument(
        "-t",
        "--timeframe",
        type=str,
        default="all",
        choices=list(get_args(TIMEFRAME)),
        help="([bold][green]bulk/semi-bulk[/][/]) timeframe to get data from (default: [green]%(default)s[/])",
    )
    main_parser.add_argument(
        "-s",
        "--sort",
        type=str,
        default="all",
        choices=list(get_args(SORT_CRITERION)),
        help="([bold][green]bulk/semi-bulk[/][/]) sort criterion (default: [green]%(default)s[/])",
    )
    main_parser.add_argument(
        "-l",
        "--limit",
        type=int,
        default=100,
        help="([bold][green]bulk/semi-bulk[/][/]) data output limit (default: [cyan]%(default)s[/])",
    )
    main_parser.add_argument(
        "--time-format",
        default="locale",
        help="determines the format of the output time (default: [green]%(default)s[/])",
        choices=["concise", "locale"],
    )
    main_parser.add_argument(
        "-e",
        "--export",
        type=str,
        help="a comma-separated list of file types to export the output to (supported: [green]csv,html,json,xml[/])",
    )
    main_parser.add_argument(
        "-u",
        "--updates",
        help="check for updates",
        action="store_true",
    )
    main_parser.add_argument(
        "-v",
        "--version",
        version=Markdown(f"Knew Karma {Version.release} {Help.copyright}"),
        action="version",
    )

    post_parser = subparsers.add_parser(
        "post",
        help="post module ([bold][green]semi-bulk[/][/])",
        description=Markdown(
            "**Post Module**: *Pull an individual post's data*",
            style="argparse.text",
        ),
        epilog=Markdown(Help.examples["post"]),
        formatter_class=RichHelpFormatter,
    )

    post_parser.add_argument("id", help="post id", type=str)
    post_parser.add_argument("subreddit", help="post source subreddit", type=str)
    post_parser.add_argument("-d", "--data", help="get post data", action="store_true")
    post_parser.add_argument(
        "-c", "--comments", help="get post comments", action="store_true"
    )

    posts_parser = subparsers.add_parser(
        "posts",
        help="posts module ([bold][green]bulk[/][/])",
        description=Markdown(
            "**Posts**: *Pull posts from various sources.*", style="argparse.text"
        ),
        epilog=Markdown(Help.examples["posts"]),
        formatter_class=RichHelpFormatter,
    )
    posts_parser.add_argument(
        "-n",
        "--new",
        help="get new posts",
        action="store_true",
    )
    posts_parser.add_argument(
        "-f",
        "--front-page",
        help="get posts from the reddit front-page",
        action="store_true",
    )
    posts_parser.add_argument(
        "-l",
        "--listing",
        default="all",
        help="get posts from a specified listing",
        choices=list(get_args(LISTING)),
    )

    search_parser = subparsers.add_parser(
        "search",
        help="search module ([bold][green]bulk[/][/])",
        description=Markdown(
            "**Search**: *Get search results from various sources.*",
            style="argparse.text",
        ),
        epilog=Markdown(Help.examples["search"]),
        formatter_class=RichHelpFormatter,
    )
    search_parser.add_argument("query", help="search query")
    search_parser.add_argument(
        "-u", "--users", help="search users", action="store_true"
    )
    search_parser.add_argument(
        "-p", "--posts", help="search posts", action="store_true"
    )
    search_parser.add_argument(
        "-s", "--subreddits", help="search subreddits", action="store_true"
    )

    subreddit_parser = subparsers.add_parser(
        "subreddit",
        help="subreddit module ([bold][green]semi-bulk[/][/])",
        description=Markdown(
            "**Subreddit**: *Pull a specified subreddit's data.*",
            style="argparse.text",
        ),
        epilog=Markdown(Help.examples["subreddit"]),
        formatter_class=RichHelpFormatter,
    )
    subreddit_parser.add_argument(
        "subreddit",
        help="subreddit name",
    )
    subreddit_parser.add_argument(
        "-p",
        "--profile",
        help="get a subreddit's profile",
        action="store_true",
    )
    subreddit_parser.add_argument(
        "-s",
        "--search",
        help="get a subreddit's posts that contain the specified keyword",
        type=str,
    )
    subreddit_parser.add_argument(
        "-pp",
        "--posts",
        help="get a subreddit's posts",
        action="store_true",
    )
    subreddit_parser.add_argument(
        "-wp",
        "--wiki-page",
        dest="wiki_page",
        help="get a subreddit's specified wiki page data",
        metavar="WIKI_PAGE",
    )
    subreddit_parser.add_argument(
        "-wps",
        "--wiki-pages",
        dest="wiki_pages",
        help="get a subreddit's wiki pages",
        action="store_true",
    )

    subreddits_parser = subparsers.add_parser(
        "subreddits",
        help="subreddits module ([bold][green]bulk[/][/])",
        description=Markdown(
            "**Subreddits**: *Pull subreddits from various sources.*",
            style="argparse.text",
        ),
        epilog=Markdown(Help.examples["subreddits"]),
        formatter_class=RichHelpFormatter,
    )
    subreddits_parser.add_argument(
        "-a",
        "--all",
        help="get all subreddits",
        action="store_true",
    )
    subreddits_parser.add_argument(
        "-d",
        "--default",
        help="get default subreddits",
        action="store_true",
    )
    subreddits_parser.add_argument(
        "-n",
        "--new",
        help="get new subreddits",
        action="store_true",
    )
    subreddits_parser.add_argument(
        "-p",
        "--popular",
        help="get popular subreddits",
        action="store_true",
    )

    user_parser = subparsers.add_parser(
        "user",
        help="user module ([bold][green]semi-bulk[/][/])",
        description=Markdown(
            "**User**: *Pull a specified user's data.*", style="argparse.text"
        ),
        epilog=Markdown(Help.examples["user"]),
        formatter_class=RichHelpFormatter,
    )
    user_parser.add_argument("username", help="username")
    user_parser.add_argument(
        "-p",
        "--profile",
        help="get a user's profile",
        action="store_true",
    )
    user_parser.add_argument(
        "-c",
        "--comments",
        help="get a user's comments",
        action="store_true",
    )

    user_parser.add_argument(
        "-o",
        "--overview",
        help="get a user's most recent comment activity",
        action="store_true",
    )
    user_parser.add_argument(
        "-pp",
        "--posts",
        action="store_true",
        help="get a user's posts",
    )
    user_parser.add_argument(
        "-sp",
        "--search-posts",
        dest="search_posts",
        help="get a user's posts that contain the specified keyword",
        type=str,
    )
    user_parser.add_argument(
        "-sc",
        "--search-comments",
        dest="search_comments",
        help="get a user's comments that contain the specified keyword",
        type=str,
    )
    user_parser.add_argument(
        "-mc",
        "--moderated-subreddits",
        dest="moderated_subreddits",
        help="get subreddits moderated by the user",
        action="store_true",
    )
    user_parser.add_argument(
        "-tc",
        "--top-subreddits",
        dest="top_subreddits",
        metavar="TOP_N",
        type=int,
        help="get a user's top n subreddits based on subreddit frequency in n posts",
    )

    return main_parser


async def call_functions(args: argparse.Namespace, function_mapping: dict):
    """
    Calls command-line arguments' functions based on user-input.

    :param args: Argparse namespace object  containing parsed command-line arguments.
    :type args: argparse.Namespace
    :param function_mapping: Mapping of command-line commands to their respective functions
    :type function_mapping: dict
    """

    async with aiohttp.ClientSession() as session:
        if args.updates:
            await Api().get_updates(session=session)
        if args.module:

            mode_action = function_mapping.get(args.module)
            directory: str = ""
            for action, function in mode_action:
                arg_is_present: bool = False
                if getattr(args, action, False):
                    arg_is_present = True

                    if args.export:
                        output_dir: str = os.path.expanduser(
                            os.path.join("~", "knewkarma-output")
                        )
                        # Create path to main directory in which target data files will be exported
                        directory = os.path.join(output_dir, args.module, action)

                        # Create file directories for supported data file types
                        pathfinder(
                            directories=[
                                os.path.join(directory, "csv"),
                                os.path.join(directory, "html"),
                                os.path.join(directory, "json"),
                                os.path.join(directory, "xml"),
                            ]
                        )

                    function_data = await function(session=session)
                    if function_data:
                        dataframe = create_dataframe(data=function_data)

                        # Print the DataFrame, excluding the 'raw_data' column if it exists
                        console.log(dataframe)

                        if args.export:
                            export_dataframe(
                                dataframe=dataframe,
                                filename=filename_timestamp(),
                                directory=directory,
                                formats=args.export.split(","),
                            )

                            # Show exported files
                            tree = Tree(
                                f":open_file_folder: [bold]{directory}[/]",
                                guide_style="bold bright_blue",
                            )
                            show_exported_files(tree=tree, directory=directory)
                            console.log(tree)

                    break

            if not arg_is_present:
                console.log(
                    f"knewkarma [underline]{args.module}[/]: missing one or more expected argument(s)"
                )


def start_cli():
    """
    Main entrypoint for the Knew Karma command-line interface.
    """
    parser = create_parser()
    args: argparse = parser.parse_args()

    start_time: datetime = datetime.now()

    limit: int = args.limit
    sort = args.sort
    timeframe = args.timeframe
    time_format = args.time_format

    search_query = args.query if hasattr(args, "query") else None

    user = User(
        username=args.username if hasattr(args, "username") else None,
        time_format=time_format,
    )
    search = Search(time_format=time_format)
    subreddit = Subreddit(
        subreddit=args.subreddit if hasattr(args, "subreddit") else None,
        time_format=time_format,
    )
    subreddits = Subreddits(time_format=time_format)
    post = Post(
        post_id=args.id if hasattr(args, "id") else None,
        subreddit=args.subreddit if hasattr(args, "subreddit") else None,
        time_format=time_format,
    )
    posts = Posts(time_format=time_format)

    function_mapping: dict = {
        "user": [
            ("profile", lambda session: user.profile(session=session)),
            (
                "posts",
                lambda session: user.posts(
                    limit=limit, sort=sort, timeframe=timeframe, session=session
                ),
            ),
            (
                "comments",
                lambda session: user.comments(
                    limit=limit, sort=sort, timeframe=timeframe, session=session
                ),
            ),
            ("overview", lambda session: user.overview(limit=limit, session=session)),
            (
                "moderated_subreddits",
                lambda session: user.moderated_subreddits(session=session),
            ),
            (
                "search_posts",
                lambda session: user.search_posts(
                    keyword=args.search_posts,
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
            (
                "search_comments",
                lambda session: user.search_comments(
                    keyword=args.search_comments,
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
            (
                "top_subreddits",
                lambda session: user.top_subreddits(
                    top_n=(
                        args.top_subreddits if hasattr(args, "top_subreddits") else None
                    ),
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
        ],
        "subreddit": [
            ("profile", lambda session: subreddit.profile(session=session)),
            (
                "posts",
                lambda session: subreddit.posts(
                    limit=limit, sort=sort, timeframe=timeframe, session=session
                ),
            ),
            (
                "search",
                lambda session: subreddit.search(
                    keyword=args.search,
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
            ("wiki_pages", lambda session: subreddit.wiki_pages(session=session)),
            (
                "wiki_page",
                lambda session: subreddit.wiki_page(
                    page=args.wiki_page if hasattr(args, "wiki_page") else None,
                    session=session,
                ),
            ),
        ],
        "subreddits": [
            ("all", lambda session: subreddits.all(limit=limit, session=session)),
            (
                "default",
                lambda session: subreddits.default(limit=limit, session=session),
            ),
            ("new", lambda session: subreddits.new(limit=limit, session=session)),
            (
                "popular",
                lambda session: subreddits.popular(limit=limit, session=session),
            ),
        ],
        "post": [
            ("profile", lambda session: post.profile(session=session)),
            (
                "comments",
                lambda session: post.comments(
                    limit=limit,
                    sort=sort,
                    session=session,
                ),
            ),
        ],
        "posts": [
            ("new", lambda session: posts.new(limit=limit, sort=sort, session=session)),
            (
                "front_page",
                lambda session: posts.front_page(
                    limit=limit, sort=sort, timeframe=timeframe, session=session
                ),
            ),
            (
                "listing",
                lambda session: posts.listing(
                    listings_name=args.listing,
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
        ],
        "search": [
            (
                "users",
                lambda session: search.users(
                    query=search_query, limit=limit, session=session
                ),
            ),
            (
                "subreddits",
                lambda session: search.subreddits(
                    query=search_query, limit=limit, session=session
                ),
            ),
            (
                "posts",
                lambda session: search.posts(
                    query=search_query,
                    limit=limit,
                    sort=sort,
                    timeframe=timeframe,
                    session=session,
                ),
            ),
        ],
    }

    if args:
        print_banner()
        try:
            start_time: datetime = datetime.now()
            asyncio.run(call_functions(args=args, function_mapping=function_mapping))
        except KeyboardInterrupt:
            console.log("[yellow]✘[/] User interruption detected ([yellow]Ctrl+C[/])")
        finally:
            elapsed_time = datetime.now() - start_time
            console.log(
                f"[green]✔[/] Done! {elapsed_time.total_seconds():.2f} seconds elapsed."
            )
    else:
        parser.print_usage()



import asyncio
import locale
import os
from datetime import datetime, timezone
from typing import Union, Literal

import pandas as pd
from rich.console import Console
from rich.status import Status
from rich.table import Table
from rich.tree import Tree


def print_banner():
    """
    Prints the projects banner and system information.
    """
    console.log(
        f"""
┓┏┓         ┓┏┓
┃┫ ┏┓┏┓┓┏┏  ┃┫ ┏┓┏┓┏┳┓┏┓
┛┗┛┛┗┗ ┗┻┛  ┛┗┛┗┻┛ ┛┗┗┗┻"""
    )
    system_info()


def system_info():
    """
    Displays system information in a table format.
    """
    import getpass
    import platform
    import sys

    import psutil

    from .version import Version

    table = Table(show_header=False, show_edge=False, highlight=True)
    table.add_column("header", style="dim")
    table.add_column("header")

    # https://www.geeksforgeeks.org/getting-the-time-since-os-startup-using-python/
    if os.name == "nt":
        # ctypes required for using GetTickCount64()
        import ctypes

        # getting the library in which GetTickCount64() resides
        lib = ctypes.windll.kernel32

        # calling the function and storing the return value
        t = lib.GetTickCount64()

        # since the time is in milliseconds i.e. 1000 * seconds
        # therefore truncating the value
        t = int(str(t)[:-3])

        # extracting hours, minutes, seconds & days from t
        # variable (which stores total time in seconds)
        minutes, seconds = divmod(t, 60)
        hours, minutes = divmod(minutes, 60)
        days, hours = divmod(hours, 24)

        system_uptime = (
            "just now"
            if days == hours == minutes == seconds == 0
            else f"{days} days, {hours:02} hours, {minutes:02} minutes, and {seconds:02} seconds since boot"
        )
    else:
        system_uptime = _time_since(
            timestamp=int(psutil.boot_time()), suffix="since boot"
        )

    table.add_row("Knew Karma", Version.full)
    table.add_row("Python", sys.version)
    table.add_row("Username", getpass.getuser())
    table.add_row("System", f"{platform.system()} {platform.version()}")
    table.add_row(
        "CPU", f"{psutil.cpu_count(logical=True)} cores, {platform.processor()}"
    )
    table.add_row(
        "Disk",
        f"{psutil.disk_usage('/').free / (1024**3):.2f} GB free"
        f" / {psutil.disk_usage('/').total / (1024**3):.2f} GB",
    )
    table.add_row(
        "Memory",
        f"{psutil.virtual_memory().available / (1024**3):.2f} GB free"
        f" / {psutil.virtual_memory().total / (1024**3):.2f} GB",
    )

    table.add_row(
        "Uptime",
        system_uptime,
    )
    console.print(table)


def get_status(status_message: str) -> Status:
    """
    Creates and returns a Status object initialized with a specific status message.

    :param status_message: A message to initialise the Status with.
    :type status_message: str
    :return: A configured rich.status.Status object ready to be used in a context manager.
    :rtype: rich.status.Status
    """
    with Status(status=status_message, spinner="dots2", console=console) as status:
        return status


async def countdown_timer(
    status: Status, duration: int, current_count: int, limit: int
):
    """
    Handles the countdown during the asynchronous pagination, updating the status bar with the remaining time.

    :param status: The rich.status.Status object used to display the countdown.
    :type status: rich.status.Status
    :param duration: The duration for which to run the countdown.
    :type duration: int
    :param current_count: Current number of items fetched.
    :type current_count: int
    :param limit: Total number of items to fetch.
    :type limit: int
    """
    for remaining in range(duration, 0, -1):
        status.update(
            f"[cyan]{current_count}[/] of [cyan]{limit}[/] "
            f"items fetched so far. Resuming in [cyan]{remaining}[/]"
            f" {'second' if remaining <= 1 else 'seconds'}[yellow]...[/]"
        )
        await asyncio.sleep(1)  # Sleep for one second as part of countdown


def _timestamp_to_locale(timestamp: float) -> str:
    """
    Converts a unix timestamp to a localized datetime string based on the system's locale.

    :param timestamp: Unix timestamp to convert.
    :type timestamp: float
    :return: A localized datetime string from the converted timestamp.
    :rtype: str
    """
    # Set the locale to the user's system default
    locale.setlocale(locale.LC_TIME, "")

    # Convert timestamp to a timezone-aware datetime object in UTC
    utc_object = datetime.fromtimestamp(timestamp, timezone.utc)

    local_object = utc_object.astimezone()

    # Format the datetime object according to the locale's conventions
    return local_object.strftime("%x, %X")


def _time_since(timestamp: int, suffix: str = "ago") -> str:
    """
    Convert a Unix timestamp into a human-readable time difference.

    :param timestamp: A Unix timestamp.
    :type timestamp: int
    :return: A string representing the time difference from now.
    :rtype: str
    """
    import time

    # Convert the current time to a Unix timestamp
    now = int(time.time())

    # Calculate the difference in seconds
    diff = now - timestamp

    # Define the time thresholds in seconds
    minute = 60
    hour = 60 * minute
    day = 24 * hour
    week = 7 * day
    month = 30 * day
    year = 12 * month

    # Determine the time unit and value
    if diff < minute:
        count = diff
        label = "seconds" if int(count) > 1 else "second"  # seconds
    elif diff < hour:
        count = diff // minute
        label = "minutes" if int(count) > 1 else "minute"  # minutes
    elif diff < day:
        count = diff // hour
        label = "hours" if int(count) > 1 else "hour"  # hours
    elif diff < week:
        count = diff // day
        label = "days" if int(count) > 1 else "day"
    elif diff < month:
        count = diff // week
        label = "weeks" if int(count) > 1 else "week"
    elif diff < year:
        count = diff // month
        label = "months" if int(count) > 1 else "month"
    else:
        count = diff // year
        label = "years" if int(count) > 1 else "year"

    return "just now" if int(count) == 0 else f"{int(count)} {label} {suffix}"


def timestamp_to_readable(
    timestamp: float, time_format: Literal["concise", "locale"] = "locale"
) -> str:
    """
    Converts a Unix timestamp into a more readable format based on the specified `time_format`.
    The function supports converting the timestamp into either a localized datetime string or a concise
    human-readable time difference (e.g., "3 hours ago").

    :param timestamp: The Unix timestamp to be converted.
    :type timestamp: float
    :param time_format: Determines the format of the output time. Use "concise" for a human-readable
                        time difference, or "locale" for a localized datetime string. Defaults to "locale".
    :type time_format: Literal["concise", "locale"]
    :return: A string representing the formatted time. The format is determined by the `time_format` parameter.
    :rtype: str
    :raises ValueError: If `time_format` is not one of the expected values ("concise" or "locale").
    """
    if time_format == "concise":
        return _time_since(timestamp=int(timestamp))
    elif time_format == "locale":
        return _timestamp_to_locale(timestamp=timestamp)
    else:
        raise ValueError(
            f"Unknown time format {time_format}. Expected `concise` or `locale`."
        )


def filename_timestamp() -> str:
    """
    Generates a timestamp string suitable for file naming, based on the current date and time.
    The format of the timestamp is adapted based on the operating system.

    :return: The formatted timestamp as a string. The format is "%d-%B-%Y-%I-%M-%S%p" for Windows
             and "%d-%B-%Y-%I:%M:%S%p" for non-Windows systems.
    :rtype: str

    Example
    -------
    - Windows: "20-July-1969-08-17-45PM"
    - Non-Windows: "20-July-1969-08:17:45PM"
    """
    now = datetime.now()
    return (
        now.strftime("%d-%B-%Y-%I-%M-%S%p")
        if os.name == "nt"
        else now.strftime("%d-%B-%Y-%I:%M:%S%p")
    )


def pathfinder(directories: list[str]):
    """
    Creates directories in knewkarma-output directory of the user's home folder.

    :param directories: A list of file directories to create.
    :type directories: list[str]
    """
    for directory in directories:
        os.makedirs(directory, exist_ok=True)


def create_dataframe(
    data: Union[dict, list[dict], list[tuple]],
):
    """
    Converts provided data into a pandas DataFrame.

    :param data: Data to be converted.
    :type data: Union[dict, list[dict], list[str]]
    :return: A pandas DataFrame constructed from the provided data. Excludes any 'raw_data'
             column from the dataframe.
    :rtype: pd.DataFrame
    """

    if isinstance(data, dict):
        # Transform each attribute of the object into a dictionary entry
        data = [{"key": key, "value": value} for key, value in data.items()]

    # Convert a list of objects (Comment, Community, Post, PreviewCommunity, User) to a list of dictionaries
    elif isinstance(data, list) and all(
        isinstance(item, (dict, tuple)) for item in data
    ):
        # Each object in the list is converted to its dictionary representation
        data = [item for item in data]

    # Set pandas display option to show all rows
    pd.set_option("display.max_rows", None)

    # Create a DataFrame from the processed data
    dataframe = pd.DataFrame(data)

    return dataframe


def show_exported_files(tree: Tree, directory: str, base_path: str = ""):
    for item in sorted(os.listdir(directory)):
        path = os.path.join(directory, item)
        if os.path.isdir(path):
            branch = tree.add(f":open_file_folder: {item}", guide_style="blue")
            show_exported_files(branch, path, os.path.join(base_path, item))
        else:
            filepath: str = os.path.join(directory, path, item, path)
            tree.add(f":page_facing_up: [italic][link file://{filepath}]{item}[/]")


def export_dataframe(
    dataframe: pd.DataFrame,
    filename: str,
    directory: str,
    formats: list[Literal["csv", "html", "json", "xml"]],
):
    """
    Exports a Pandas dataframe to specified file formats.

    :param dataframe: Pandas dataframe to export.
    :type dataframe: pandas.DataFrame
    :param filename: Name of the file to which the dataframe will be exported.
    :type filename: str
    :param directory: Directory to which the dataframe files will be saved.
    :type directory: str
    :param formats: A list of file formats to which the data will be exported.
    :type formats: list[Literal]
    """
    file_mapping: dict = {
        "csv": lambda: dataframe.to_csv(
            os.path.join(directory, "csv", f"{filename}.csv"), encoding="utf-8"
        ),
        "html": lambda: dataframe.to_html(
            os.path.join(directory, "html", f"{filename}.html"),
            escape=False,
            encoding="utf-8",
        ),
        "json": lambda: dataframe.to_json(
            os.path.join(directory, "json", f"{filename}.json"),
            orient="records",
            lines=True,
            force_ascii=False,
            indent=4,
        ),
        "xml": lambda: dataframe.to_xml(
            os.path.join(directory, "xml", f"{filename}.xml"),
            parser="etree",
            encoding="utf-8",
        ),
    }

    for file_format in formats:
        if file_format in file_mapping:
            filepath: str = os.path.join(
                directory, file_format, f"{filename}.{file_format}"
            )
            file_mapping.get(file_format)()
            console.log(
                f"[green]✔[/] {os.path.getsize(filepath)} bytes written to [link file://{filepath}]{filepath}"
            )
        else:
            console.log(f"Unsupported file format: {file_format}")


console = Console(color_system="auto", log_time=False)



from datetime import date


class Help:
    """
    Container for documentation-related data for Knew Karma.

    Attributes:
          author (str): A markdown-formatted string of author name and url.
          copyright (str): Copyright notice for the current year, including the author's details.
          summary (str): A brief description of Knew Karma as a tool for Reddit data analysis.
          description (str): A full description of Knew Karma as a CLI, Library,
                and GUI program for Reddit data analysis.
          examples (dict): Usage examples for different operations within Knew Karma.
    """

    author: str = "[Richard Mwewa](https://rly0nheart.github.io)"
    copyright: str = (
        f"© Copyright 2023-{date.today().year} {author}. All rights reserved."
    )

    summary: str = f"**Knew Karma**: *A Reddit Data Analysis Toolkit* — by {author}"
    description: str = """
**Knew Karma** (/nuː ‘kɑːrmə/) is a **Reddit** Data Analysis Toolkit designed to provide an extensive range of
functionalities for exploring and analysing Reddit data. It includes a **Command-Line Interface (CLI)**, an
**Application Programming Interface (API)** to enable an easy integration in other Python Projects and a **Graphical
User
Interface (GUI)** for Windows machines, making it adaptable for various user preferences.
"""

    examples: dict = {
        "user": """
The **User** module is used to pull a specified user's data including profile, comments, posts, 
searching posts that contain the specified keyword, or even searching comments that contain the specified keyword.

# Examples
## Get user profile
```
knewkarma user  AutoModerator --profile
```

## Get user comments
```
knewkarma user  AutoModerator --comments
```

## Get user posts
```
knewkarma user  AutoModerator --posts
```

## Get user's most recent comment activity
```
knewkarma user AutoModerator --overview
```

## Get a user's posts that contain the specified keyword
```
knewkarma user AutoModerator --search-posts rules
```

## Get a user's comment that contain the specified keyword
```
knewkarma user AutoModerator --search-comments banned
```

## Get subreddits moderated by user
```
knewkarma user TheRealKSi --moderated-subreddits
```

## Get user's top n subreddits based on subreddit frequency in n posts
```
knewkarma --limit 500 user TheRealKSi --top-subreddits 10
```
""",
        "subreddit": """
The **Subreddit** module is used to pull a specified subreddit's data including profile, posts, wiki pages,
 or even searching posts that have the specified keyword.

# Examples
## Get subreddit profile
```
knewkarma subreddit MachineLearning --profile
```

## Get subreddit posts
```
knewkarma subreddit MachineLearning --posts
```

## Search a subreddit for posts that contain a specified keyword
```
knewkarma subreddit MachineLearning --search "artificial intelligence"
```

## Get subreddit's wiki pages
```
knewkarma subreddit MachineLearning --wiki-pages
```

## Get subreddit's specified wiki page data
```
knewkarma subreddit MachineLearning --wiki-page config/description
```
    """,
        "subreddits": """
The **Subreddits** module is used to pull subreddits from various sources such as new, default, or popular.

# Examples
## Get all subreddits
```
knewkarma subreddits --all
```

## Get default subreddits
```
knewkarma subreddits --default
```

## Get new subreddits
```
knewkarma subreddits --new
```

## Get popular subreddits
```
knewkarma subreddits --popular
```
    """,
        "post": """
The **Post** module is used to pull an individual post's data including its comments, 
provided the post's **id** and source **subreddit** are provided.

# Examples
### Get a post's data
```
knewkarma post 12csg48 OSINT --data
```

### Get a post's comments
```
knewkarma post 12csg48 OSINT --comments
```
""",
        "posts": """
The **Posts** is used to pull posts from sources like the front page, a listing, or new posts.

# Examples
## Get new posts
```
knewkarma posts --new
```

## Get posts from Reddit Front-Page
```
knewkarma posts --front-page
```

## Get posts from specified listing
```
knewkarma posts --listing best
```
    """,
        "search": """
The **Search** module is used for search/discovery of targets in users, subreddits, and posts.

# Examples
## Search users
```
knewkarma search john --users
```

## Search subreddits
```
knewkarma search ask --subreddits
```

## Search posts
```
knewkarma search covid-19 --posts
```
    """,
    }



class Version:
    """
    Represents Knew Karma's version in a structured format, providing access
    to the major, minor, and patch components, as well as composite representations
    like the full version string and the release string.

    Attributes:
        major (str): The major version number, indicating major changes and milestones.
        minor (str): The minor version number, representing minor improvements and significant fixes.
        patch (str): The patch version number, typically used for small fixes and updates.
        full (str): A composite string representing the full version, combining major, minor, and patch.
        release (str): A string representing the release version, combining only the major and minor numbers.
    """

    major: str = "4"
    minor: str = "2"
    patch: str = "3"
    full: str = f"{major}.{minor}.{patch}"
    release: str = f"{major}.{minor}"



from random import randint
from sys import version as python_version
from typing import Union, Literal

import aiohttp
from rich.markdown import Markdown

from ._utilities import console, countdown_timer, get_status
from .version import Version

SORT_CRITERION = Literal[
    "controversial",
    "new",
    "top",
    "best",
    "hot",
    "rising",
]

LISTING = Literal["best", "controversial", "popular", "rising"]
TIMEFRAME = Literal["hour", "day", "week", "month", "year"]
TIME_FORMAT = Literal["concise", "locale"]


class Api:
    """Represents the Knew Karma API and provides methods for getting various data from the Reddit API."""

    def __init__(self):
        self._base_reddit_endpoint: str = "https://www.reddit.com"
        self._user_data_endpoint: str = f"{self._base_reddit_endpoint}/u"
        self._users_data_endpoint: str = f"{self._base_reddit_endpoint}/users"
        self.subreddit_data_endpoint: str = f"{self._base_reddit_endpoint}/r"
        self._subreddits_data_endpoint: str = f"{self._base_reddit_endpoint}/subreddits"
        self._github_release_endpoint: str = (
            "https://api.github.com/repos/bellingcat/knewkarma/releases/latest"
        )

    @staticmethod
    async def get_data(
        session: aiohttp.ClientSession, endpoint: str
    ) -> Union[dict, list]:
        """
        Asynchronously fetches JSON data from a given API endpoint.

        :param session: aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :param endpoint: The API endpoint to fetch data from.
        :type endpoint: str
        :return: Returns JSON data as a dictionary or list. Returns an empty dict if fetching fails.
        :rtype: Union[dict, list]
        """
        try:
            async with session.get(
                endpoint,
                headers={
                    "User-Agent": f"Knew-Karma/{Version.release} "
                    f"(Python {python_version}; +https://knewkarma.readthedocs.io)"
                },
            ) as response:
                if response.status == 200:
                    return await response.json()
                else:
                    error_message = await response.json()
                    console.log(
                        f"[red]✘[/] An API error occurred: [red]{error_message}[/]"
                    )
                    return {}

        except aiohttp.ClientConnectionError as error:
            console.log(f"[red]✘[/] An HTTP error occurred: [red]{error}[/]")
            return {}
        except Exception as error:
            console.log(f"[red]✘[/] An unknown error occurred: [red]{error}[/]")
            return {}

    @staticmethod
    def _process_response(
        response_data: Union[dict, list], valid_key: str = None
    ) -> Union[dict, list]:
        """
        Processes and validates the API response data.

        If it's a dictionary and a valid_key is provided,
        checks for the presence of the key in the response dictionary.

        If it's a list, it ensures the list is not empty.

        :param response_data: The API response data to validate, which should be a dictionary or list.
        :type response_data: Union[dict, list]
        :param valid_key: The key to check for in the data if it's a dictionary.
        :type valid_key: str
        :return: The original data if valid, or an empty dictionary or list if invalid.
        :rtype: Union[dict, list]
        """
        if isinstance(response_data, dict):
            if valid_key:
                return response_data if valid_key in response_data else {}
            else:
                return response_data
        elif isinstance(response_data, list):
            return response_data if response_data else []
        else:
            console.log(
                f"[yellow]✘[/] Unknown data type ({response_data}: {type(response_data)}), expected a list or dict."
            )

    async def get_updates(self, session: aiohttp.ClientSession):
        """
        Asynchronously gets and compares the current program version with the remote version.

        Assumes version format: major.minor.patch.prefix

        :param session: aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        """
        with get_status(status_message=f"Checking for updates..."):
            # Make a GET request to PyPI to get the project's latest release.
            response: dict = await self.get_data(
                endpoint=self._github_release_endpoint, session=session
            )
            release: dict = self._process_response(
                response_data=response, valid_key="tag_name"
            )

            if release:
                remote_version: str = release.get("tag_name")
                markup_release_notes: str = release.get("body")
                markdown_release_notes = Markdown(markup=markup_release_notes)

                # Splitting the version strings into components
                remote_parts: list = remote_version.split(".")

                update_message: str = (
                    f"[underline]%s[/] update ([underline]{remote_version}[/]) available"
                )

                # Check for differences in version parts
                if remote_parts[0] != Version.major:
                    update_message = update_message % "MAJOR"

                elif remote_parts[1] != Version.minor:
                    update_message = update_message % "MINOR"

                elif remote_parts[2] != Version.patch:
                    update_message = update_message % "PATCH"

                if update_message:
                    console.log(update_message)
                    console.log(markdown_release_notes)

    async def get_profile(
        self,
        profile_source: str,
        profile_type: Literal["user", "subreddit"],
        session: aiohttp.ClientSession,
    ) -> dict:
        """
        Asynchronously fetches a profile from the specified source.

        :param profile_source: Source to get profile data from.
        :type profile_source: str
        :param profile_type: The type of profile that is to be fetched.
        :type profile_type: str
        :param session: aiohttp session to use for the request.
        :return: A dictionary object containing profile data from the specified source.
        :rtype: dict
        """
        with get_status(
            status_message=f"Initialising [underline]single data[/] retrieval..."
        ):
            # Use a dictionary for direct mapping
            profile_mapping: dict = {
                "user": f"{self._user_data_endpoint}/{profile_source}/about.json",
                "subreddit": f"{self.subreddit_data_endpoint}/{profile_source}/about.json",
            }

            # Get the endpoint directly from the dictionary
            endpoint: str = profile_mapping.get(profile_type, "")

            profile_data = await self.get_data(endpoint=endpoint, session=session)

        return self._process_response(
            response_data=profile_data.get("data", {}), valid_key="created_utc"
        )

    async def _paginate(
        self,
        session: aiohttp.ClientSession,
        endpoint: str,
        process_func: callable,
        limit: int,
    ) -> list[dict]:
        """
        Asynchronously fetches and processes data in a paginated manner
        from a specified endpoint until the specified limit
        of items is reached or there are no more items to fetch. It uses a specified processing function
        to handle the data from each request, ensuring no duplicates are returned.

        :param session: An Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :param endpoint: Endpoint to get paginated data from.
        :type endpoint: str
        :param process_func: A callable function for processing response data.
        :type process_func: callable
        :param limit: Maximum number of results to return.
        :type limit: int
        :return: A list of dict objects, each containing paginated data.
        :rtype: list[dict]
        """
        all_items = []
        last_item_id = None
        with get_status(
            status_message="Initialising [underline]bulk data[/] retrieval..."
        ) as status:
            while len(all_items) < limit:
                paginated_endpoint = (
                    f"{endpoint}&after={last_item_id}&count={len(all_items)}"
                    if last_item_id
                    else endpoint
                )

                response = await self.get_data(
                    session=session, endpoint=paginated_endpoint
                )
                items = response.get("data", {}).get("children", [])

                if not items:
                    break

                processed_items = process_func(response_data=items)
                items_to_limit = limit - len(all_items)
                all_items.extend(processed_items[:items_to_limit])

                last_item_id = response.get("data").get("after")

                if len(all_items) == limit:
                    break

                sleep_delay: int = randint(1, 10)

                await countdown_timer(
                    status=status,
                    duration=sleep_delay,
                    current_count=len(all_items),
                    limit=limit,
                )

        return all_items

    async def get_posts(
        self,
        session: aiohttp.ClientSession,
        limit: int,
        posts_type: Literal[
            "new_posts",
            "front_page_posts",
            "listing_posts",
            "subreddit_posts",
            "user_posts",
            "user_overview",
            "user_comments",
        ],
        posts_source: str = None,
        timeframe: TIMEFRAME = "all",
        sort: SORT_CRITERION = "all",
    ) -> list[dict]:
        """
        Asynchronously gets a specified number of posts, with a specified sorting criterion, from the specified source.

        :param session: aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :param limit: Maximum number of posts to get.
        :type limit: int
        :param posts_type: Type of posts to be fetched.
        :type posts_type: str
        :param posts_source: Source from where posts will be fetched.
        :type posts_source: str
        :param sort: Posts' sort criterion.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str

        :return: A list of dictionaries, each containing data of a post.
        :rtype: list[dict]
        """
        source_map = {
            "new_posts": f"{self._base_reddit_endpoint}/new.json",
            "front_page_posts": f"{self._base_reddit_endpoint}/.json",
            "listing_posts": f"{self.subreddit_data_endpoint}/{posts_source}.json?",
            "subreddit_posts": f"{self.subreddit_data_endpoint}/{posts_source}.json",
            "user_posts": f"{self._user_data_endpoint}/{posts_source}/submitted.json",
            "user_overview": f"{self._user_data_endpoint}/{posts_source}/overview.json",
            "user_comments": f"{self._user_data_endpoint}/{posts_source}/comments.json",
        }

        endpoint = source_map.get(posts_type, "")
        if posts_type == "new_posts":
            endpoint += f"?limit={limit}&sort={sort}&raw_json=1"
        else:
            endpoint += f"?limit={limit}&sort={sort}&t={timeframe}&raw_json=1"

        posts: list[dict] = await self._paginate(
            session=session,
            endpoint=endpoint,
            process_func=self._process_response,
            limit=limit,
        )

        return posts

    async def get_search_results(
        self,
        session: aiohttp.ClientSession,
        search_type: Literal["users", "subreddits", "posts"],
        query: str,
        limit: int,
        sort: SORT_CRITERION = "all",
        timeframe: TIMEFRAME = "all",
    ) -> list[dict]:
        """
        Asynchronously searches from a specified results type that match the specified query.

        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        :param search_type: Type of results to get.
        :type search_type: str
        :param query: Search query.
        :type query: str
        :param limit: Maximum number of results to get.
        :type limit: int
        :param sort: Posts' sort criterion.
        :type sort: str
        :param timeframe: Timeframe from which to get posts.
        :type timeframe: str
        """
        search_mapping: dict = {
            "users": f"{self._users_data_endpoint}/search.json",
            "subreddits": f"{self._subreddits_data_endpoint}/search.json",
            "posts": f"{self._base_reddit_endpoint}/search.json",
        }

        endpoint = search_mapping.get(search_type, "")
        if search_type == "posts":
            endpoint += f"?q={query}&limit={limit}&sort={sort}&t={timeframe}"
        else:
            endpoint += f"?q={query}&limit={limit}"

        search_results: list[dict] = await self._paginate(
            session=session,
            endpoint=endpoint,
            process_func=self._process_response,
            limit=limit,
        )

        return search_results

    async def get_subreddits(
        self,
        subreddits_type: Literal["all", "default", "new", "popular"],
        limit: int,
        session: aiohttp.ClientSession,
    ) -> list[dict]:
        """
        Asynchronously gets the specified type of subreddits.

        :param subreddits_type: Type of subreddits to get.
        :type subreddits_type: str
        :param limit: Maximum number of subreddits to return.
        :type limit: int
        :param session: Aiohttp session to use for the request.
        :type session: aiohttp.ClientSession
        """
        subreddits_mapping: dict = {
            "all": f"{self._subreddits_data_endpoint}.json",
            "default": f"{self._subreddits_data_endpoint}/default.json",
            "new": f"{self._subreddits_data_endpoint}/new.json",
            "popular": f"{self._subreddits_data_endpoint}/popular.json",
        }

        endpoint = subreddits_mapping.get(subreddits_type, "")
        endpoint += f"?limit={limit}"

        subreddits: list[dict] = await self._paginate(
            session=session,
            endpoint=endpoint,
            process_func=self._process_response,
            limit=limit,
        )

        return subreddits
